<?php

namespace Drupal\commerce_stripe_webhook_event;

use Drupal\Core\Entity\EntityInterface;
use Stripe\Event;
use Symfony\Component\HttpFoundation\Request;

/**
 * Provides functionality for webhook events.
 */
class WebhookEvent {

  public const STATUS_UNPROCESSED = 0;

  public const STATUS_SUCCEEDED = 1;

  public const STATUS_FAILED = 2;

  public const STATUS_SKIPPED = 3;

  /**
   * Gets an array of statuses.
   *
   * @return array
   *   An array of statuses.
   */
  public static function getStatuses(): array {
    return [
      self::STATUS_UNPROCESSED => t('Unprocessed'),
      self::STATUS_SUCCEEDED => t('Succeeded'),
      self::STATUS_FAILED => t('Failed'),
      self::STATUS_SKIPPED => t('Skipped'),
    ];
  }

  /**
   * Gathers a list of uniquely defined webhook event types.
   *
   * @return array
   *   List of uniquely defined webhook event types.
   */
  public static function getEventTypes(): array {
    return \Drupal::database()
      ->query('SELECT DISTINCT([type]) FROM {commerce_stripe_webhook_event} ORDER BY [type]')
      ->fetchAllKeyed(0, 0);
  }

  /**
   * Creates a list of webhook event filters that can be applied.
   *
   * @return array
   *   Associative array of filters. The top-level keys are used as the form
   *   element names for the filters, and the values are arrays with the
   *   following elements:
   *   - title: Title of the filter.
   *   - where: The filter condition.
   *   - options: Array of options for the select list for the filter.
   */
  public static function getFilters(): array {
    $filters = [];

    foreach (self::getEventTypes() as $type) {
      $types[$type] = $type;
    }

    if (!empty($types)) {
      $filters['type'] = [
        'title' => t('Type'),
        'where' => 'w.type = ?',
        'options' => $types,
      ];
    }

    $filters['status'] = [
      'title' => t('Status'),
      'where' => 'w.status = ?',
      'options' => self::getStatuses(),
    ];

    return $filters;
  }

  /**
   * Insert a new webhook event.
   *
   * @param \Symfony\Component\HttpFoundation\Request $request
   *   The request.
   * @param \Stripe\Event $webhook_event
   *   The webhook event.
   * @param string $stripe_signature
   *   The stripe signature.
   *
   * @return int|null
   *   The ID of the webhook event or NULL if it already exists.
   *
   * @throws \Exception
   */
  public static function insert(Request $request, Event $webhook_event, string $stripe_signature): ?int {
    $payload = $request->getContent();

    /** @var \Stripe\ApiResource $event_object */
    $event_object = $webhook_event->data->object;

    $database = \Drupal::database();
    if ($database->select('commerce_stripe_webhook_event')->condition('stripe_event_id', $webhook_event->id)->countQuery()->execute()->fetchField() > 0) {
      // If we have received this event previously, we can ignore it.
      // We might consider logging this, to determine if we are receiving
      // an inordinate amount of repeat events.
      return NULL;
    }
    $webhook_event_id = $database
      ->insert('commerce_stripe_webhook_event')
      ->fields([
        'stripe_event_id' => $webhook_event->id,
        'type' => $webhook_event->type,
        'status' => self::STATUS_UNPROCESSED,
        'payload' => $payload,
        'ip' => $request->getClientIp() ?? '',
        'processed' => 0,
        'received' => \Drupal::time()->getRequestTime(),
        'signature' => $stripe_signature,
        'stripe_object_type' => $event_object->object,
        'stripe_object_id' => $event_object->id,
      ])
      ->execute();
    if ($webhook_event_id > 0) {
      return $webhook_event_id;
    }
    throw new \RuntimeException('Unable to store event');
  }

  /**
   * Update the webhook event status.
   *
   * @param int $webhook_event_id
   *   The id of the webhook event.
   * @param int $status
   *   The status.
   * @param string|null $reason
   *   The reason for the status.
   * @param \Drupal\Core\Entity\EntityInterface|null $entity
   *   The related entity.
   */
  public static function updateStatus(int $webhook_event_id, int $status, ?string $reason, ?EntityInterface $entity = NULL): void {
    $database = \Drupal::database();
    $fields = [
      'status' => $status,
      'processed' => ($status !== self::STATUS_UNPROCESSED) ? \Drupal::time()->getCurrentTime() : 0,
      'reason' => $reason ?? '',
    ];
    if ($entity !== NULL && $entity->id() > 0) {
      $fields['entity_type'] = $entity->getEntityTypeId();
      $fields['entity_id'] = $entity->id();
    }
    $database->update('commerce_stripe_webhook_event')
      ->fields($fields)
      ->condition('webhook_event_id', $webhook_event_id)
      ->execute();
  }

}
